﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using HarmonyLib;
using UnityEngine.SceneManagement;
using UnityEngine.EventSystems;
using UnityEngine.UI;
using TMPro;

[HarmonyPatch(typeof(ItemInstance), "Uses", MethodType.Setter)]
class Patch_SetItemUses
{
    static void Prefix(ItemInstance __instance, ref int __0)
    {
        if (BetterInventory.Singleton.player.Inventory.GetSelectedHotbarSlot().itemInstance == null)
        {
            return;
        }
        if(__0 == 0 && BetterInventory.Singleton.player.Inventory.GetItemCount(__instance.baseItem.UniqueName) > 1 && __instance.settings_usable.IsUsable())
        {
            foreach(Slot s in BetterInventory.Singleton.player.Inventory.allSlots)
            {
                if(s.itemInstance != null && s.itemInstance.UniqueName == __instance.UniqueName && s != BetterInventory.Singleton.player.Inventory.GetSelectedHotbarSlot())
                {
                    BetterInventory.Singleton.player.Inventory.GetSelectedHotbarSlot().SetItem(s.itemInstance);
                    s.Reset();
                    break;
                }
            }
        }
    }
}


public class BetterInventory : Mod
{
    private static BetterInventory singleton;
    public static BetterInventory Singleton 
    {
        get
        {
            return singleton;
        }
        set
        {
            if(singleton == null)
            {
                singleton = value;
            }
            //else
            //{
            //    Destroy(value.gameObject);
            //}
        }
    }
    public Network_Player player;
    public static AssetBundle asset;
    // Bad name, was supposed to only be the sort button, evolved to be the stand in for every instantiate.
    GameObject SortButton;
    GameObject SearchBar;
    GameObject SearchButton;
    DontSortController dontSortController;
    Harmony harmony;

    //objectsToDestroy gets ever GO that is instantiated so it can be destroyed when the mod unloads. So there are no artifacts.
    List<GameObject> objectsToDestroy = new List<GameObject>();
    List<Item_Base> itemsPreSorted = new List<Item_Base>(); 

    // Quick hack so I can easily add more modes if needed
    private enum SortMode
    {
        ID,
        Alph,
        Group
    }
    bool hasStarted = false;
    public IEnumerator Start()
    {
        AssetBundleCreateRequest request = AssetBundle.LoadFromMemoryAsync(GetEmbeddedFileBytes("nei.assets"));                     // This assetpack only contains basic graphics. Graphics are applied by code.
        yield return request;
        asset = request.assetBundle;
        hasStarted = true;
    }

    public IEnumerator StartMod()
    {
        print("Starting Better Inventories");
        Singleton = this;
        harmony = new Harmony("com.example.patch");
        harmony.PatchAll();

        if (SceneManager.GetActiveScene().name != "MainScene")
        {
            yield break;
        }

        GameObject go = null;
        player = RAPI.GetLocalPlayer();
        go = asset.LoadAsset<GameObject>("Button");                                                                                 // A default button to instantiate.
        GameObject dontSortAmount = asset.LoadAsset<GameObject>("DontSortAmount");                                                  // The button setting the unsorted amount of cells
        GameObject searchBarPrefab = asset.LoadAsset<GameObject>("SearchInput");
        TMP_FontAsset chineseRocksFont = asset.LoadAsset<TMP_FontAsset>("ChineseRocks SDF");                                        // The ChineseRocks Font for TMP.
        Font chineseRocksF = asset.LoadAsset<Font>("ChineseRocks");                                                                 // The ChineseRocks Font for normal Texts. This could also be extracted from the gtamefiles, might refactor for less assetpack size.
        Transform slotParent = GameObject.Find("Canvases/_CanvasGame_New/InventoryParent/Inventory_Player/SlotParent").transform;   // The parent with the Layout group for inventory slots.
        Sprite defaultSquareBg = GameObject.Find("Canvases/_CanvasGame_New/InventoryParent/Inventory_Player/SlotParent/Slot_Inventory").GetComponent<Image>().sprite;   // This is the default background used for item slots. Caught from the game so it is not needed in the assetpack.
        itemsPreSorted = ItemManager.GetAllItems();                                                                                 // The pre sorted list, this basically comes in kind of a group sort as of 22/03/2022

        // Instantiates the dontSort button and applies graphics. Again "SortButon" has devolved to a wrong variable name.
        SortButton = Instantiate(dontSortAmount, slotParent);
        dontSortController = SortButton.AddComponent<DontSortController>();
        SortButton.transform.SetSiblingIndex(1);
        objectsToDestroy.Add(SortButton);
        SortButton.GetComponentsInChildren<Button>().ToList().ForEach(bu => 
        {
            bu.image.sprite = defaultSquareBg;
            Text t = bu.gameObject.GetComponentInChildren<Text>();
            t.font = chineseRocksF;
            t.color = new Color(0.9137255f, 0.8352941f, 0.6745098f, 1f);
            t.resizeTextForBestFit = true;
        });
        // Instantiates the buttons that are used to sort on index 1-4 (hence beginning at 1, aending at 4)
        for (int i = 1; i < 5; i++)
        {
            // Instantiates a button, sets some parameteres like color, background image, text font, auto sizing. Color values should be the same as other in game UIs.
            SortButton = Instantiate(go, slotParent);
            SortButton.GetComponent<Image>().sprite = defaultSquareBg;
            TextMeshProUGUI t = SortButton.GetComponentInChildren<TextMeshProUGUI>();
            t.enableAutoSizing = true;
            t.color = new Color(0.9137255f, 0.8352941f, 0.6745098f, 1f);
            t.font = chineseRocksFont;
            Shadow sh = SortButton.AddComponent<Shadow>();
            sh.effectColor = new Color(0.372549f, 0.2235294f, 0.1254902f, 1f);
            sh.effectDistance = new Vector2(2, -2);
            sh.useGraphicAlpha = true;

            TextMeshProUGUI buttonText = SortButton.GetComponentInChildren<TextMeshProUGUI>();
            Button btn = SortButton.GetComponent<Button>();
            switch (i)
            {
                case 1:
                    btn.onClick.AddListener(onSortByIDButtonClick);
                    buttonText.text = "Sort\nBy ID";
                    break;
                case 2:
                    btn.onClick.AddListener(onSortAlphButtonClick);
                    buttonText.text = "Sort\nA-Z";
                    break;
                case 3:
                    btn.onClick.AddListener(onSortByGroupButtonClick);
                    buttonText.text = "Sort\nBy Group";
                    break;
                case 4:
                    btn.onClick.AddListener(onSearchButtonClicked);
                    buttonText.text = "Search";
                    SearchButton = SortButton;
                    break;
                default:
                    break;
            }
            SortButton.transform.SetSiblingIndex(i + 1);                // Set index up so it appears in front of the LayoutControl
            objectsToDestroy.Add(SortButton);
        }
        // Spawn the search bar on top of the buttons and disable it, so it covers the buttons when called and there is no need to take further action.
        SearchBar = Instantiate(searchBarPrefab, slotParent);
        objectsToDestroy.Add(SearchBar);
        SearchBar.SetActive(false);
        SearchBar.GetComponent<Image>().sprite = defaultSquareBg;
        SearchBar.GetComponentsInChildren<TMP_Text>().ToList().ForEach(t =>
        {
            t.enableAutoSizing = true;
            t.color = new Color(0.9137255f, 0.8352941f, 0.6745098f, 1f);
            t.font = chineseRocksFont;
        });
        SearchBar.GetComponent<TMP_InputField>().onValueChanged.AddListener(onSearchValueChanged);


        Debug.Log("BetterInventories has been loaded!");
    }

    private void onSortByGroupButtonClick()
    {
        Sort(SlotsToSort(), SortMode.Group);
        if (player.Inventory.secondInventory != null)
        {
            Sort(player.Inventory.secondInventory.allSlots, SortMode.Group);
        }
    }

    private void onSortByIDButtonClick()
    {
        onSortAlphButtonClick();            // For some reason we need to sort alphabetically first else it doesnt work .-.
        Sort(SlotsToSort(), SortMode.ID);
        if (player.Inventory.secondInventory != null)
        {
            Sort(player.Inventory.secondInventory.allSlots, SortMode.ID);
        }
    }

    private void onSortAlphButtonClick()
    {
        Sort(SlotsToSort(), SortMode.Alph);
        if (player.Inventory.secondInventory != null)
        {
            Sort(player.Inventory.secondInventory.allSlots, SortMode.Alph);
        }
    }

    private List<Slot> SlotsToSort()
    {
        // Get all of the inventory slots that are not in the hotbar.
        List<Slot> nonHotbar = player.Inventory.allSlots.GetRange(player.Inventory.hotslotCount, player.Inventory.allSlots.Count - player.Inventory.hotslotCount);
        nonHotbar.RemoveRange(0, dontSortController.DontSortAmount);

        return nonHotbar;
    }

    private void onSearchValueChanged(string value)
    {
        List<Slot> slots = player.Inventory.allSlots.GetRange(player.Inventory.hotslotCount, player.Inventory.allSlots.Count - player.Inventory.hotslotCount);
        if (player.Inventory.secondInventory != null)
        {
            slots.AddRange(player.Inventory.secondInventory.allSlots);
        }
        // Iterate over the inventory, if the item in slot x has a name that contains the search term, let it be highlighted, if not, turn down its opacity, maybe change bg color to dark?
        foreach (Slot s in slots)
        {
            Image i = s.transform.Find("ItemImage").GetComponent<Image>();
            if (s.itemInstance == null)
            {
                continue;
            }
            if (s.itemInstance.UniqueName.ToLower().Contains(value.ToLower()))
            {
                i.color = new Color(i.color.r, i.color.g, i.color.b, 1f);
            }
            else
            {
                i.color = new Color(i.color.r, i.color.g, i.color.b, .25f);
            }
        }
    }

    private void onSearchButtonClicked()
    {
        // If the Searchbutton has been clicked and the bar is still active, deactivate it, else, activate it. Mainly just activating the search bar, setting canvas, and altering text.
        TMP_InputField searchInput = SearchBar.GetComponent<TMP_InputField>();
        if (SearchBar.activeSelf)
        {
            SearchButton.GetComponentInChildren<TextMeshProUGUI>().text = "Search";
            CanvasHelper.ActiveMenu = MenuType.Inventory;
            searchInput.text = string.Empty;
        }
        else
        {
            SearchButton.GetComponentInChildren<TextMeshProUGUI>().text = "Stop\nSearch";
            CanvasHelper.ActiveMenu = MenuType.TextWriter;
            searchInput.Select();
            searchInput.ActivateInputField();
        }
        SearchBar.SetActive(!SearchBar.activeSelf);             // Toggle state
    }

    bool hasInit = false;
    public void Update()
    {
        if(SceneManager.GetActiveScene().name == "MainScene" && !hasInit && hasStarted)
        {
            StartCoroutine(StartMod());
            hasInit = true;
        }
        else if (SceneManager.GetActiveScene().name != "MainScene" && hasInit)
        {
            hasInit = false;
        }
        if (CanvasHelper.ActiveMenu == MenuType.TextWriter && SearchBar != null && SearchBar.activeSelf && Input.GetKeyDown(KeyCode.Escape))
        {
            onSearchButtonClicked();
        }
    }

    #region Credits: Traxan, wazupwiop (modified by me)
    private void Sort(List<Slot> nonHotbar, SortMode mode)
    {
        // Get a temporary list of all the used slots in the inventory.
        List<ItemInstance> tempInstances = new List<ItemInstance>();

        foreach (Slot slot in nonHotbar)
        {
            // Grab its amount.
            if (!slot.IsEmpty)
            {
                tempInstances.Add(slot.itemInstance.Clone());
            }
            // Reset the slot.
            slot.Reset();
        }

        switch (mode)
        {
            case SortMode.ID:
                tempInstances = tempInstances.OrderBy(s => s.UniqueIndex).ToList();
                break;
            case SortMode.Alph:
                tempInstances.Sort(new ItemComparer());
                break;
            case SortMode.Group:
                List<ItemInstance> ttInstances = new List<ItemInstance>();
                foreach (Item_Base item in itemsPreSorted)
                {
                    List<ItemInstance> tt = tempInstances.Where(ti => ti.baseItem.UniqueIndex == item.UniqueIndex).ToList();
                    if (tt.Count() != 0)
                    {
                        print(tt.FirstOrDefault().UniqueName);
                        print(tt.Count());
                        ttInstances.AddRange(tt);
                    }
                }
                tempInstances = ttInstances;
                break;
        }

        // Do some assignments to sort it in the inventory
        var i = 0;
        foreach (ItemInstance instance in tempInstances)
        {
            Slot currSlot = nonHotbar[i];
            currSlot.SetItem(instance);
            i++;
        }

        // Now we compress with i starting at 1
        for (i = 1; i < nonHotbar.Count; i++)
        {
            Slot currSlot = nonHotbar[i];
            if (!currSlot.IsEmpty)
            {
                Slot toSlot = this.FindSuitableSlot(nonHotbar, i, currSlot.GetItemBase());
                if (toSlot != currSlot && toSlot != null)
                {
                    this.TransferItems(toSlot, currSlot);
                }
            }
        }
    }

    private void TransferItems(Slot toSlot, Slot fromSlot)
    {
        if (!fromSlot.itemInstance.settings_Inventory.Stackable || toSlot.IsEmpty)
        {
            toSlot.SetItem(fromSlot.itemInstance);
            fromSlot.Reset();
        }
        else
        {
            int maxAmount = fromSlot.itemInstance.settings_Inventory.StackSize;
            if (fromSlot.itemInstance.Amount + toSlot.itemInstance.Amount > maxAmount)
            {
                int diff = maxAmount - toSlot.itemInstance.Amount;
                fromSlot.itemInstance.Amount -= diff;
                toSlot.itemInstance.Amount += diff;
                fromSlot.RefreshComponents();
                toSlot.RefreshComponents();
            }
            else
            {
                toSlot.itemInstance.Amount += fromSlot.itemInstance.Amount;
                toSlot.RefreshComponents();
                fromSlot.Reset();
            }
        }
    }

    private Slot FindSuitableSlot(List<Slot> inv, int currIndex, Item_Base stackableItem = null)
    {
        bool flag = stackableItem != null && stackableItem.settings_Inventory.Stackable;
        for (var i = 0; i < currIndex; i++)
        {
            Slot slot2 = inv[i];
            if (slot2.IsEmpty)
            {
                return slot2;
            }
            if (flag && !slot2.StackIsFull() && !slot2.IsEmpty && slot2.itemInstance.UniqueIndex == stackableItem.UniqueIndex)
            {
                return slot2;
            }
        }
        return null;
    }

    sealed class ItemComparer : IComparer<ItemInstance>
    {
        public int Compare(ItemInstance x, ItemInstance y)
        {
            var localized = String.Compare(x.baseItem.settings_Inventory.DisplayName, y.baseItem.settings_Inventory.DisplayName, StringComparison.CurrentCultureIgnoreCase);
            if (localized != 0)
            {
                return localized;
            }
            var unique = String.Compare(x.UniqueName, y.UniqueName, StringComparison.Ordinal);
            if (unique != 0)
            {
                return unique;
            }
            return y.Amount.CompareTo(x.Amount); // Bigger stacks to the front
        }
    }
    #endregion

    public void OnModUnload()
    {
        try
        {
            foreach (GameObject go in objectsToDestroy)
            {
                Destroy(go);
            }
        }
        catch 
        {
            Debug.Log("No objects to destroy!");
        }
        try
        {
            harmony.UnpatchAll("com.example.patch");
        }
        catch 
        {
            Debug.LogWarning("Could not unpatch harmony!");
        }
        try
        {
            asset.Unload(true);
        }
        catch (Exception e)
        {
            Debug.LogError("Could not unload assets!");
            Debug.LogError(e.Message);
        }
        finally
        {
            Debug.Log("Mod TestMod has been unloaded!");
        } 
    }
}